#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This example demonstrates how to use the Measurement Error Mitigation method to
improve the accuracy of the outcomes generated by a noisy quantum computer.
"""
import qiskit
from qiskit.providers.fake_provider import FakeSantiago

import sys
sys.path.append('../..')

from QCompute import *
from qcompute_qep.measurement import InverseCorrector, LeastSquareCorrector
from qcompute_qep.utils import execute

# Set the token. You must set your VIP token in order to access the hardware.
Define.hubToken = "Token"
# Set the default number of shots
NUMBER_OF_SHOTS = 8192
# Set the default number of qubits
NUMBER_OF_QUBITS = 2

##########################################################################################
# Setup the quantum program for the Bell state.
##########################################################################################
# qp is short for "quantum program", instance of QProgram
qp = QEnv()
qp.Q.createList(NUMBER_OF_QUBITS)
H(qp.Q[0])
CX(qp.Q[0], qp.Q[1])
MeasureZ(*qp.Q.toListPair())

##########################################################################################
# Set the quantum hardware for running the Bell circuit.
##########################################################################################
# For numeric test on the ideal simulator, change qc to BackendName.LocalBaiduSim2
# qc = BackendName.LocalBaiduSim2

# For experiment on the real quantum device, change qc to BackendName.CloudBaiduQPUQian
# qc = BackendName.CloudBaiduQPUQian

# For numeric test on the noisy simulator, change qc to Qiskit's FakeSantiago
qc = qiskit.providers.aer.AerSimulator.from_backend(FakeSantiago())

##########################################################################################
# Execute the quantum program on the quantum computer and mitigate the results.
##########################################################################################
# Obtain the noisy counts
noisy_counts = execute(qp, qc, shots=NUMBER_OF_SHOTS)
# Use the measurement mitigation procedure to correct the outcomes
# You may choose other calibrators and correctors for better results.
corrector = InverseCorrector(qc, calibrator='tp', qubits=range(NUMBER_OF_QUBITS))
counts = corrector.correct(noisy_counts)

print("Noisy Counts: {}".format(noisy_counts))
print("Corrected Counts: {}".format(counts))
